from autokeras.pretrained.object_detector import ObjectDetector
from autokeras.pretrained.face_detector import FaceDetector
from autokeras.pretrained.voice_generator.voice_generator import VoiceGenerator